/*******************************************************************************
 * udpgen.C - UDP traffic generator with scripting possibilities
 *
 * $Id: udpgen.C,v 1.2 2005/04/26 11:12:31 mscheid Exp $
 *
 * Author: Matthias Scheidegger <mscheid@iam.unibe.ch>
 *
 * Copyright 2001, Matthias Scheidegger
 *
 ******************************************************************************/

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <math.h>
#include <list>
using namespace std;

#include <netinet/in.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
extern int h_errno;

#include <sys/time.h>
#include "time.H"


//-- Usage ---------------------------------------------------------------------

static char *exe_name = NULL;     // name the executable (as supplied in argv)

void usage()
{
    fprintf(stderr, "Usage: %s [-f  <scriptfile>] [-t ttl] <desthost> <destport>\n",
	    exe_name);
    exit(2);
}


//-- Types and Constants -------------------------------------------------------


struct RefPoint
{
    double time;  // relative time of point
    double bw;  // bandwidth at point
    RefPoint(double t, double b) : time(t), bw(b) {}
};


class BWFunction
{
    struct Vertex { double time, bw, area; };

    list<Vertex> vertices;
    double func_time;
    double func_area;

  public:
    BWFunction(list<RefPoint> &reference_points);
    
    double bw_at(double time);              // bandwidth at time
    double area(double begin, double end);  // integrated bandwidth between begin and end
    bool finished(double time);             // are we finished at time
};


const int pkthdr_len = 28;  // IP and UDP Header (TODO: include ethernet header)


//-- Function declarations -----------------------------------------------------

list<RefPoint> parse_script(FILE *script);
void setup_net(const char *dname, int dport, int ttl);
void start(BWFunction &bwf);


//-- Configuration -------------------------------------------------------------

int sock;
struct sockaddr_in destination;


//-- Main ----------------------------------------------------------------------

int main(int argc, char **argv)
{
    FILE *script = NULL;
    char *dhost = NULL;
    int dport = 0;
    int ttl = 0;

    exe_name = argv[0];
    argv++; argc--;
    if (argc > 6) usage();

    // parse options
    while (argc > 2) {
	if (!strcmp("-f",argv[0])) {
	    char *fname;
	    if (script) { fprintf(stderr, "double -f\n"); exit(2); }
	    if (strlen(argv[0]) > 2) fname = argv[0]+2;
	    else if (argc >= 4) fname = argv[1];
	    else usage();
	    argv++; argc--;
	    script = fopen(fname, "r");
	    if (!script) { perror("open"); exit(2); }
	    argv++; argc--;	    
	}
	else if (!strcmp("-t",argv[0])) {
	    char *str;
	    if (ttl) { fprintf(stderr, "double -t\n"); exit(2); }
	    if (strlen(argv[0]) > 2) str = argv[0]+2;
	    else if (argc >= 4) str = argv[1];
	    else usage();
	    argv++; argc--;
	    ttl = atoi(str);
	    if (ttl <= 0 || ttl > 255) usage();
	    argv++; argc--;	    
	}
	else {
	    fprintf(stderr, "Unknown option %s\n", argv[0]);
	    usage();
	}
    }

    // parse destination
    if (argc != 2) usage();
    dhost = argv[0];
    dport = atoi(argv[1]);
    if (!dport) { fprintf(stderr, "Illegal destport\n"); usage(); }

    // parse script
    if (!script) script = stdin;
    list<RefPoint> rpoints = parse_script(script);
    BWFunction bwf(rpoints);

    // create socket and start transmission
    setup_net(dhost, dport, ttl);
    start(bwf);

    return 0;
}


//-- Parsing -------------------------------------------------------------------

list<RefPoint> parse_script(FILE *script)
{
    int point;
    double time, bw;
    list<RefPoint> refpoints;  // list of reference point for bw function

    for (point=1; !feof(script); point++) {
	if (fscanf(script, "%lf %lf", &time, &bw) != 2 ) {
	    fprintf(stderr, "Syntax error reading point %i\n", point);
	    exit(2);
	}
	if (time < 0 || bw < 0) {
	    fprintf(stderr, "Syntax error reading point %i, "
		    "no negative values allowed\n", point);
	    exit(2);	    
	}
	refpoints.push_back(RefPoint(time, bw));
	fscanf(script," ");  // eat whitespace
    }
    return refpoints;
}


//-- BWFunction methods --------------------------------------------------------

BWFunction::BWFunction(list<RefPoint> &rpoints)
{
    list<RefPoint>::iterator rpi;
    list<Vertex>::iterator vi;
    Vertex v;
    double last_bw = 0;
    
    func_time = 0;
    func_area = 0;

    for (rpi=rpoints.begin(); rpi != rpoints.end(); rpi++) {
	v.time = rpi->time;
	v.bw = rpi->bw;
	v.area = rpi->time * (last_bw+rpi->bw)/2;
	func_time += v.time;
	func_area += v.area;
	vertices.push_back(v);
	last_bw = rpi->bw;
    }
}


double BWFunction::bw_at(double t)
{
    assert(t >= 0.0);

    double last_bw = 0;
    list<Vertex>::iterator iter;

    for (iter=vertices.begin(); iter != vertices.end(); iter++) {
	if (t >= iter->time) {
	    last_bw = iter->bw;
	    t -= iter->time;
	} else {
		return last_bw + (t / iter->time) * (iter->bw - last_bw);
	}
    }
    
    return last_bw;
}


double BWFunction::area(double begin, double end)
{
    assert(begin <= end);

    double area = 0;
    double last_bw = 0;
    list<Vertex>::iterator iter;

    for (iter=vertices.begin(); iter != vertices.end(); iter++) {
	if (begin >= iter->time) {
	    begin -= iter->time;
	    end -= iter->time;
	    last_bw = iter->bw;
	} else {
	    double begin_bw = last_bw + begin/iter->time*(iter->bw - last_bw);
	    area =  (iter->time - begin) * (iter->bw+begin_bw)/2;
	    break;
	}
    }
    
    if (iter == vertices.end()) {   // both behind last vertex
	return (end-begin)*vertices.back().bw;
    }

    if (end <= iter->time) {        // begin and end in same interval
	double end_bw = last_bw + end/iter->time*(iter->bw - last_bw);
	return area - ((iter->time - end) * (iter->bw+end_bw)/2);
    } else {
	end -= iter->time;
	last_bw = iter->bw;
	++iter;
    }

    for (; iter != vertices.end(); iter++) {
	if (end >= iter->time) {
	    area += iter->area;
	    end -= iter->time;
	    last_bw = iter->bw;
	}
	else {
	    double end_bw = last_bw + end/iter->time*(iter->bw - last_bw);	    
	    return area + end *(last_bw+end_bw)/2;
	}
    }
    return area + (end - vertices.back().time)*vertices.back().bw;
}


bool BWFunction::finished(double time)
{
    if (time > func_time) return true;
    return false;
}

 
//-- Networking code -----------------------------------------------------------

void pexit(const char *function)
{
    perror(function);
    exit(2);
}


void setup_net(const char *dname, int dport, int ttl)
{
    struct hostent *he;

    // fill out address
    if ((he = gethostbyname(dname)) == NULL) {
	herror("gethostbyname");
	exit(2);
    }
    bzero(&destination, sizeof(struct sockaddr_in));
    destination.sin_family = AF_INET;
    destination.sin_port = htons((short)dport);
    memcpy(&destination.sin_addr.s_addr, he->h_addr, he->h_length);

    // create socket
    if ((sock = socket(PF_INET, SOCK_DGRAM, 0)) == -1) pexit("socket");

    // set TTL if specified
    if (ttl && setsockopt(sock, IPPROTO_IP, IP_TTL, &ttl, sizeof(ttl)))
	pexit("setsockopt");
}



void start(BWFunction &bwf)
{
    const double base_ival = 0.01; // 100th sec
    const int min_pl   = 2;
    const int min_size = pkthdr_len + min_pl;
    const int max_pl   = 1400;
    const int max_size = pkthdr_len + max_pl;

    char buffer[max_pl];

    // Made payload carry a 16bit sequence number <mscheid 5.3.04>
    unsigned short *seqno = (unsigned short*)buffer;
    *seqno = 0;

    const Time st;                 // start time (init here)
    Time now(0.0);
    struct timeval tval;
    double dt;                     // delta time


    for (now=st, dt=0; !bwf.finished(dt); now.set(), dt=now-st) {
	double ival = base_ival;
	double area;
	int to_send;
	Time to_wait(0.0);

	// find suitable interval
	area = bwf.area(dt,dt+ival);

	if (area == 0) {
	    // zero bandwidth on interval. expand to length of zero stretch
	    do {
		ival += base_ival;
		area = bwf.area(dt,dt+ival);
	    }
	    while (area == 0 && !bwf.finished(dt+ival));
	    if (area != 0) ival -= base_ival;
	    else return;
	}
	else if (area < min_size) {
	    // area to small for packet. expand interval to increase area
	    do {
		ival += base_ival;
		area = bwf.area(dt,dt+ival);
	    }
	    while (area < min_size && !bwf.finished(dt+ival));
	    if (area < min_size) return;
	}
	to_send = (int)area;

	// send
	while (to_send) {
	    int send_size, nof_sends, i;

	    if (to_send > max_size) send_size = max_size; // multiple sends
	    else if (to_send < min_size) break;           // too little to send, give up
	    else send_size = to_send;                     // in between
	    nof_sends = to_send / send_size;
	    to_send -= (send_size * nof_sends);

	    for (i=0; i<nof_sends; i++) {
		now.set();
		sendto(sock, buffer, send_size-pkthdr_len, 0,
		       (struct sockaddr*)&destination, sizeof(struct sockaddr_in));
		*seqno += 1;
	    }
	}

	// wait during the rest of the interval
	now.set();
	to_wait = (dt+ival) - (now-st);
	if (0 >= to_wait) continue;
	to_wait.get_timeval(&tval);
	select(0, 0,0,0, &tval);
    }
}
